import com.sun.media.jai.codec.TIFFEncodeParam;
import com.sun.media.jai.codec.SeekableOutputStream;
import com.sun.media.jai.codec.TIFFField;
import com.sun.media.jai.codecimpl.TIFFImageDecoder;
import java.awt.image.BufferedImage;
import java.awt.image.renderable.ParameterBlock;
import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.Collection;
import javax.media.jai.JAI;
import javax.media.jai.RenderedOp;

import org.apache.fop.apps.Driver;
import org.apache.fop.apps.FOPException;
import org.apache.fop.render.awt.AWTRenderer;
import org.xml.sax.InputSource;

//test
import java.awt.Transparency;
import java.awt.RenderingHints;
import java.awt.color.ColorSpace;
import java.awt.image.ColorModel;
import java.awt.image.ComponentColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.PixelInterleavedSampleModel;
import java.awt.image.SampleModel;
import javax.media.jai.ImageLayout;

/**
 * Generates images from graphics.
 *
 * @author Andriy Palamarchuk
 */
public class FO2Image {

    /**
     * Converts RGB image to gray.
     *
     * @param colorImage the RGB image to prepare
     * @return gray image
     */
    RenderedOp convertToGray(BufferedImage colorImage){
        ParameterBlock pb = new ParameterBlock();
        pb.addSource(colorImage);

        // Create a grayscale color model.
        ColorSpace cs = ColorSpace.getInstance(ColorSpace.CS_GRAY);
        int bits[] = new int[] {8};
        ColorModel cm = new ComponentColorModel(cs, bits, false, false,
                Transparency.OPAQUE, DataBuffer.TYPE_BYTE);
        pb.add(cm);

        //Create a sample model with the same number of bands
        //as the  number of components in the colour model.
        int numBands = cs.getNumComponents();
        //Set up the band offsets for the Sample Model.
        int[] bandOffsets = new int[numBands];
        for (int i = 0; i < numBands; i++) {
            bandOffsets[i] = numBands - 1 - i;
        }

        //Create pixel and scanline strides.
        int pixelStride = numBands;
        int scanlineStride = numBands * colorImage.getWidth();

        //Create the sample model.
        SampleModel sm = new PixelInterleavedSampleModel(DataBuffer.TYPE_BYTE,
                colorImage.getWidth(), colorImage.getHeight(),
                pixelStride, scanlineStride, bandOffsets);

        // Create a tiled layout with the requested ColorModel.
        ImageLayout layout = new ImageLayout();

        layout.setTileWidth(colorImage.getTileWidth());
        layout.setTileHeight(colorImage.getTileHeight());
        layout.setColorModel(cm);
        layout.setSampleModel(sm);
        // Create RenderingHints for the ImageLayout.
        RenderingHints rh = new RenderingHints(JAI.KEY_IMAGE_LAYOUT, layout);

        //Create the grayscale image using the ColorConvert operation.
        return JAI.create("colorconvert", pb, rh);
    }

    /**
     * Returns image, ready for saving in tiff file from provided image.
     *
     * @param pageImage page image to prepare
     * @return prepared image
     */
    RenderedOp preparePage(BufferedImage pageImage) {
        ParameterBlock pb;
        RenderedOp image;

        image = convertToGray(pageImage);

        pb = new ParameterBlock();
        pb.addSource(image);
        image = JAI.create("invert", pb);

        pb = new ParameterBlock();
        pb.addSource(image);
        pb.add(5.6);
        image = JAI.create("binarize", pb);

        return JAI.create("invert", image);
    }
    
    /**
     * Generates image.
     *
     * @param format 
     */
    public void generateImage(String format, File inFile, File outFile)
            throws IOException, FOPException {
        Driver driver = new Driver(new InputSource(inFile.getPath()), null);
        AWTRenderer renderer = new AWTRenderer(null);
        driver.setRenderer(renderer);
        driver.run();

        if (format.equals("show")) {
            new ShowImage().show(renderer);
        } else if (format.equals("tiff")) {
            RenderedOp firstPageImage = null;
            Collection extraImages = new ArrayList();

            renderer.setScaleFactor(300);

            OutputStream out =
                new SeekableOutputStream(new RandomAccessFile(outFile, "rw"));
            
            TIFFEncodeParam  encodeParams = new TIFFEncodeParam();
            encodeParams.setCompression(TIFFEncodeParam.COMPRESSION_GROUP4);

            //write to file x and y resolution
            TIFFField extraFields[] = new TIFFField[2];
            long xResolution[][] = new long[1][2];
            xResolution[0][0] = 200;
            xResolution[0][1] = 1;
            extraFields[0] = new TIFFField(TIFFImageDecoder.TIFF_X_RESOLUTION,
                    TIFFField.TIFF_RATIONAL, 1, xResolution);

            long yResolution[][] = new long[1][2];
            yResolution[0][0] = 200;
            yResolution[0][1] = 1;
            extraFields[1] = new TIFFField(TIFFImageDecoder.TIFF_Y_RESOLUTION,
                    TIFFField.TIFF_RATIONAL, 1, yResolution);
            encodeParams.setExtraFields(extraFields);

            //create page images
            for (int page = 0; page < renderer.getPageCount(); page++) {
                renderer.render(page);
                if (page == 0) {
                    firstPageImage =
                            preparePage(renderer.getLastRenderedPage());
                } else {
                    extraImages.add(
                            preparePage(renderer.getLastRenderedPage()));
                }
            }
            encodeParams.setExtraImages(extraImages.iterator());

            JAI.create("encode", firstPageImage, out, "TIFF", encodeParams);
            out.close();

        } else if (format.equals("gif")) {
            System.out.println("Format '" + format + "' is not implemented");
            System.exit(1);
        } else {
            System.out.println("Format '" + format + "' is not supported");
            System.exit(1);
        }
    }

    /**
     * Runs the image generator.
     * See the source code of the method for command-line parameters
     * description.
     */
    public static void main(String args[])
        throws IOException, FOPException {
        FO2Image fo2Image = new FO2Image();
        
        if (!(args.length == 3 || (args[0].equals("show") && args.length == 2))) {
            System.out.println("Incorrect format");
            System.out.println("The call:\n\tFO2Image <output format> <fo file> [<image file>]");
            System.out.println("\nWhere:");
            System.out.println(
                    "\t<output format> - one of following: show, tiff, gif. " +
                    "The image is displayed if the format is \"show\", the image " +
                    "file is not necessary");
            System.exit(1);
        }

        String format = args[0];
        File in = new File(args[1]);
        File out = new File(args[2]);
        if (!in.exists()) {
            System.out.println("Input file " + in + " does not exist!");
            System.exit(1);
        }
        fo2Image.generateImage(format, in, out);
    }
}
